//--------------------------------------------------------------------------------------------------
//
// full_duplex_serial_ts.c
//
// A full duplex serial driver using Propeller task switch (JMPRET).
// Runs in a Propeller COG as native code.
//
// Copyright (c) 2011 Michael Rychlik.
// MIT Licensed (see at end of file for exact terms)
//
// History:
//
// 2011-10-08    v0.1  Initial version.
//
//--------------------------------------------------------------------------------------------------

#include "full_duplex_serial_ts.h"

// For Cog regs (_OUTA, _INA, _DIRA..)
#include "cog.h"

#define taskswitch(newfunc) __builtin_propeller_taskswitch((newfunc)) 


// Pointer to mail box in HUB
static _COGMEM volatile struct fds_mailbox* pm;

// Serial port config
static _COGMEM int txPin;
static _COGMEM int rxPin;
static _COGMEM int bitcycles;
static _COGMEM int halfBitcycles;

// tx_Thread variables
static _COGMEM int txWaitCycles;
static _COGMEM int txShiftReg;
static _COGMEM int txBitCnt;

static _COGMEM volatile void* txTask;
static _COGMEM volatile void* rxTask;


// The UART transmitter function.
_NATIVE static void tx_thread()
{
    // We loop forever here.
    while(1)
    {
        // Wait while the TX FIFO is empty

//      PT_YIELD_UNTIL(pt, pm->txHead != pm->txTail);
        while (pm->txHead == pm->txTail)   // WARNING! This while loop fails when it confuses Z flag during task switch
        {
            txTask = taskswitch(rxTask);
        }

        // Load the tx shift register from tx FIFO and add stop bits.
        txShiftReg = pm->txBuffer[(pm->txTail)++ & BUFFER_MASK]; 

        // Output the start bit
        _OUTA = 0;
        
	// Get the next bit time
        txWaitCycles = (int)_CNT + bitcycles;

        for (txBitCnt = 0; txBitCnt < 8; txBitCnt++)
        {
//          PT_YIELD_UNTIL(pt, (int)_CNT - txWaitCycles >= 0); // Wait for the next bit time
            while ((int)_CNT - txWaitCycles < 0)
            {
                txTask = taskswitch(rxTask);
            }		

            _OUTA = (txShiftReg & 1) << txPin;                // Output a bit 
            txWaitCycles += bitcycles;                        // Get next bit time
            txShiftReg >>= 1;                                 // Get next bit into position
        }
        // Output the stop bit 

//      PT_YIELD_UNTIL(pt, (int)_CNT - txWaitCycles >= 0);     // Wait for the next bit time
        while ((int)_CNT - txWaitCycles < 0)
	{
            txTask = taskswitch(rxTask);
	}

        _OUTA = 1 << txPin; 

        // Wait for end  of stop  bit time 
        txWaitCycles += bitcycles;                            // Get next bit time

//      PT_YIELD_UNTIL(pt, (int)_CNT - txWaitCycles >= 0);
        while ((int)_CNT - txWaitCycles < 0)
	{
            txTask = taskswitch(rxTask);
	}
    }
}

// rx_thread variables
static _COGMEM int rxWaitCycles;
static _COGMEM int rxShiftReg;
static _COGMEM int rxBitCnt;
static _COGMEM int bit;
static _COGMEM int temp;

// The UART reciever function.
_NATIVE static void rx_thread()
{
    // We loop forever here.
    while(1)
    {
        rxShiftReg = 0;
        // Wait until a start bit is detected

//      PT_YIELD_UNTIL(pt, ((_INA >> rxPin) & 1) == 0);
        while (((_INA >> rxPin) & 1) != 0)
        {
            rxTask = taskswitch(txTask);
        }

        // Wait until halfway through start bit
        rxWaitCycles = (int)_CNT + halfBitcycles - 200;

//      PT_YIELD_UNTIL(pt, (int)_CNT - rxWaitCycles >= 0);
        while ((int)_CNT - rxWaitCycles < 0)
        {
            rxTask = taskswitch(txTask);
        }

        rxWaitCycles = (int)_CNT + bitcycles - 200;     // Get next bit time
        
        for (rxBitCnt = 0; rxBitCnt <= 7; rxBitCnt++)
        { 
//          PT_YIELD_UNTIL(pt, (int)_CNT - rxWaitCycles  >= 0);   // Wait for middle of bit time 
            while ((int)_CNT - rxWaitCycles < 0)
            {
                rxTask = taskswitch(txTask);
            }


            bit = (_INA >> rxPin) & 1;                            // Read the bit 
            rxWaitCycles = (int)_CNT + bitcycles;                 // Get next bit time               

            // WARNING! the << operator causes the stack to be used
            rxShiftReg |= bit << rxBitCnt;                        // and place in shift register
        } 

	// Write the recieved char to the RX FIFO
	if ((pm->rxHead - pm->rxTail) < BUFFER_SIZE - 1)
        {
                // Not full: Put char in buffer.
                pm->rxBuffer[pm->rxHead++ & BUFFER_MASK] = (char)rxShiftReg;
	}
	rxWaitCycles -= 200;

        // Wait into center of Stop bit
//      PT_YIELD_UNTIL(pt, (int)_CNT - rxWaitCycles >= 0); 
        while ((int)_CNT - rxWaitCycles < 0)
        {
            rxTask = taskswitch(txTask);
        }
    }
}


_NATIVE void main (volatile struct fds_mailbox *m)
{
    // Copy mail box pointer for other functions use.
    pm = m;

    // Set up serial port
    rxPin = pm->rxPin;
    txPin = pm->txPin;
    bitcycles = 80000000 / pm->baudRate;  //<HEATER>Shouldn't CLKFREQ be defined somewhere visible here?</HEATER>
    halfBitcycles = bitcycles / 2;
    _OUTA = (unsigned)1 << txPin;
    _DIRA = (unsigned)1 << txPin;

    rxTask = rx_thread;
    txTask = tx_thread;

    // Start the reciever thread
    // which will in turn run the transmiter thread as a coroutine.
    rx_thread ();
}

//--------------------------------------------------------------------------------------------------
// TERMS OF USE: MIT License
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software,
// and to permit persons to whom the Software is furnished to do so,
// subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//--------------------------------------------------------------------------------------------------
